import { test, rejects, flow, modules } from '../runner.js'

for (const module of modules('ensure-jarm-without-iss-fails')) {
  test.serial(rejects(flow()), module, 'JWT "iss" (issuer) claim missing')
}
